'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /* eslint no-param-reassign: ["error", { "props": false }] */

var _queue = require('./queue');

var _queue2 = _interopRequireDefault(_queue);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Graph = function () {
  function Graph() {
    _classCallCheck(this, Graph);

    this.vertices = [];
    this.adjList = {};
  }

  _createClass(Graph, [{
    key: 'addVertex',
    value: function addVertex(v) {
      this.vertices.push(v);
      this.adjList[v] = [];
    }
  }, {
    key: 'addEdge',
    value: function addEdge(v, w) {
      this.adjList[v].push(w);
    }
  }, {
    key: 'toString',
    value: function toString() {
      var s = '';
      for (var i = 0; i < this.vertices.length; i += 1) {
        s += this.vertices[i];
        s += ' -> ';
        var neighbors = this.adjList[this.vertices[i]];
        for (var j = 0; j < neighbors.length; j += 1) {
          s += neighbors[j];
          if (j !== neighbors.length - 1) {
            s += ' ';
          }
        }
        s += '\n';
      }
      return s;
    }

    /**
     * Breath-first search (BFS), first widely, then deeply.
     * start traversing the graph from the first specified vertex,
     * will visit all its neighbors first, then the next layer (neighbor's neighbor)
     * @method bfs
     * @param {Object} v - the first vertex to start the bfs
     * @param {Function} callback - callback function when visiting the vertex
     * @return {Object} there are 2 property in the return object:
     *    distances - degree of each vertex in the path
     *    predecessors - predecessor of the vertex
     */

  }, {
    key: 'bfs',
    value: function bfs(v, callback) {
      var visitingQueue = new _queue2.default();
      var visited = {};
      var distances = {};
      var predecessors = {};

      if (this.vertices.indexOf(v) >= 0) {
        visitingQueue.enqueue(v);
        visited[v] = 'inprogress';
        predecessors[v] = null;
        distances[v] = 0;
        while (!visitingQueue.isEmpty()) {
          var u = visitingQueue.dequeue();
          var neighbors = this.adjList[u];
          for (var i = 0; i < neighbors.length; i += 1) {
            var w = neighbors[i];
            if (!visited[w]) {
              visitingQueue.enqueue(w);
              visited[w] = 'inprogress';
              distances[w] = distances[u] + 1;
              predecessors[w] = u;
            }
          }
          visited[u] = 'done';
          if (callback) {
            callback(u);
          }
        }
      }
      return {
        distances: distances,
        predecessors: predecessors
      };
    }

    /**
     * Breath-first search (BFS), first widely, then deeply.
     * start traversing the graph from the first specified vertex,
     * will visit all its neighbors first, then the next layer (neighbor's neighbor)
     * @method bfs
     * @param {Function} callback - callback function when visiting the vertex
     * @return {Array} shortest path in array format. Empty array means no shortest path
     */

  }, {
    key: 'getShortestPath',
    value: function getShortestPath(sourceVertex, destVertex) {
      if (!{}.hasOwnProperty.call(this.adjList, sourceVertex) || !{}.hasOwnProperty.call(this.adjList, destVertex)) {
        // vertex not exist
        return [];
      }

      // use bfs to get the predecessors
      var path = [];

      var _bfs = this.bfs(sourceVertex, function (vertex) {
        path.push(vertex);
      }),
          predecessors = _bfs.predecessors;

      // check if there is path to the destVertex


      if (path.indexOf(destVertex) === -1) {
        return [];
      }

      // generate the shortest path
      var shortestPath = [];

      var v = destVertex;
      while (v !== sourceVertex) {
        shortestPath.push(v);
        v = predecessors[v];
      }
      shortestPath.push(sourceVertex);
      return shortestPath.reverse();
    }

    /**
     * Depth-first search (DFS), first deeply, then widely.
     * start traversing the graph from the first specified vertex,
     * will follow the path until last vertex of the path is visited,
     * then will backtrack the next path
     * @method dfs
     * @param {Object} v - the first vertex to start the dfs
     * @param {Function} callback - callback function when visiting the vertex
     * @return {Object} there are 3 property in the return object:
     *    discovery - the discovery time
     *    finished - the time marked as finish
     *    predecessors - predecessor of the vertex
     */

  }, {
    key: 'dfs',
    value: function dfs(vertex, callback) {
      var visited = {};
      var discovery = {};
      var finished = {};
      var predecessors = {};
      var counter = 0;

      if (this.vertices.indexOf(vertex) >= 0) {
        predecessors[vertex] = null;
        this.dfsVisit(vertex, visited, discovery, finished, predecessors, counter, callback);
      }

      return {
        discovery: discovery,
        finished: finished,
        predecessors: predecessors
      };
    }

    /**
     * The recursive function to DFS traverse the vertices
     * @method dfsVisit
     * @param {Object} v - the current visiting vertex
     * @param {Object} visited - dictionary for marking the node visited or not
     * @param {Object} discovery - dictionary for vertex discovery time
     * @param {Object} finished - dictionary for vertex finish visit time
     * @param {Integer} counter - the current step
     * @param {Function} callback - callback function when visiting the vertex
     * @return {Integer} The total step of the node visit or finish
     */

  }, {
    key: 'dfsVisit',
    value: function dfsVisit(v, visited, discovery, finished, predecessors, counter, callback) {
      visited[v] = true;
      var time = counter + 1;
      discovery[v] = time;
      // console.log(`discovered ${v}`);
      if (callback) {
        callback(v);
      }
      var neighbors = this.adjList[v];
      for (var i = 0; i < neighbors.length; i += 1) {
        var u = neighbors[i];
        if (!visited[u]) {
          predecessors[u] = v;
          time = this.dfsVisit(u, visited, discovery, finished, predecessors, time, callback);
        }
      }
      time += 1;
      finished[v] = time;
      // console.log(`explored ${v}`);
      return time;
    }
  }]);

  return Graph;
}();

module.exports = Graph;